import re

from .basic import BasicImplementation
from antelope import BackgroundInterface, ProductFlow, ExteriorFlow, EntityNotFound, comp_dir
from antelope_core.contexts import Context


class NonStaticBackground(Exception):
    pass


def search_skip(entity, search):
    if search is None:
        return False
    return not bool(re.search(search, str(entity), flags=re.IGNORECASE))


class BackgroundImplementation(BasicImplementation, BackgroundInterface):
    """
    The default Background Implementation exposes an ordinary inventory database as a collection of LCI results.
    Because it does not perform any ordering, there is no way to distinguish between foreground and background
    elements in a database using the proxy. It is thus inconsistent for the same resource to implement both
    inventory and [proxy] background interfaces.
    """
    def __init__(self, *args, **kwargs):
        super(BackgroundImplementation, self).__init__(*args, **kwargs)

        self._index = None

    def check_bg(self, **kwargs):
        return True

    def setup_bm(self, index=None):
        """
        Requires an index interface or catalog query <-- preferred
        :param index:
        :return:
        """
        if self._index is None:
            if index is None:
                self._index = self._archive.make_interface('index')
            else:
                self._index = index

    def _ensure_ref_flow(self, ref_flow):
        if ref_flow is not None:
            if isinstance(ref_flow, str) or isinstance(ref_flow, int):
                ref_flow = self._archive.retrieve_or_fetch_entity(ref_flow)
        return ref_flow

    def foreground_flows(self, search=None, **kwargs):
        """
        No foreground flows in the proxy background
        :param search:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def background_flows(self, search=None, **kwargs):
        """
        all process reference flows are background flows
        :param search:
        :param kwargs:
        :return:
        """
        self.check_bg()
        for p in self._index.processes():
            for rx in p.references():
                if search_skip(p, search):
                    continue
                yield ProductFlow(self._archive.ref, rx.flow, rx.direction, p, None)

    def exterior_flows(self, direction=None, search=None, **kwargs):
        """
        Exterior flows are all flows that do not have interior terminations (i.e. not found in the index targets)
        Since contexts are still in limbo, we need a default directionality (or some way to establish directionality
        for compartments..) but for now let's just use default 'output' for all exterior flows
        :param direction:
        :param search:
        :param kwargs:
        :return:
        """
        self.check_bg()
        for f in self._index.flows():
            if search_skip(f, search):
                continue
            try:
                next(self._index.targets(f.external_ref, direction=direction))
            except StopIteration:
                cx = self._index.get_context(f.context)
                dir = comp_dir(cx.sense)
                '''
                if self.is_elementary(f):
                    yield ExteriorFlow(self._archive.ref, f, 'Output', f['Compartment'])
                else:
                    yield ExteriorFlow(self._archive.ref, f, 'Output', None)
                '''
                yield ExteriorFlow(self._archive.ref, f, dir, cx)

    def consumers(self, process, ref_flow=None, **kwargs):
        """
        Not supported for trivial backgrounds
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def dependencies(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have no dependencies
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def emissions(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have only exterior flows. Emissions are the exterior flows with elementary
        contexts
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in self.lci(process, ref_flow=ref_flow, **kwargs):
            if isinstance(i.termination, Context):
                if i.termination.elementary:
                    yield i

    def cutoffs(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have only exterior flows. Emissions are the exterior flows with non-elementary
        contexts
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in self.lci(process, ref_flow=ref_flow, **kwargs):
            if isinstance(i.termination, Context):
                if i.termination.elementary:
                    continue
            yield i

    def foreground(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        ref_flow = self._ensure_ref_flow(ref_flow)
        p = self._index.get(process)
        yield p.reference(ref_flow)  # should be just one exchange

    def is_in_scc(self, process, ref_flow=None, **kwargs):
        """
        Distinction between is_in_background and is_in_scc will reveal the proxy nature of the interface
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        return False  # proxy has no knowledge of SCCs

    def is_in_background(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        try:
            self._index.get(process)
        except EntityNotFound:
            return False
        return True

    def ad(self, process, ref_flow=None, **kwargs):
        for i in []:
            yield i

    def bf(self, process, ref_flow=None, **kwargs):
        for i in []:
            yield i

    def lci(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        ref_flow = self._ensure_ref_flow(ref_flow)
        p = self._index.get(process)
        for x in p.inventory(ref_flow=ref_flow):
            yield x

    def sys_lci(self, node, demand, **kwargs):
        raise NotImplementedError

    def bg_lcia(self, process, query_qty, ref_flow=None, **kwargs):
        p = self._archive.retrieve_or_fetch_entity(process)
        lci = self.lci(p, ref_flow=ref_flow)
        res = query_qty.do_lcia(lci, locale=p['SpatialScope'], **kwargs)
        """
        if self.privacy > 0:
            return res.aggregate('*', entity_id=p.link)
        """
        return res



import unittest
from synonym_dict.example_compartments.test_compartments import CompartmentContainer, InconsistentLineage
from ..contexts import Context, ContextManager, InconsistentSense
from antelope.interfaces.iindex import InvalidSense
from ..lcia_engine.lcia_engine import DEFAULT_CONTEXTS


class ContextTest(CompartmentContainer.CompartmentTest):
    def test_sense(self):
        c = Context('emissions to air', sense='sink')
        self.assertEqual(c.sense, 'Sink')
        with self.assertRaises(InvalidSense):
            Context('emissions to Mars', sense='extraterrestrial')
        with self.assertRaises(InconsistentSense):
            Context('resources from urban air', parent=c, sense='source')

    def test_elementary(self):
        c = Context('eMissions', sense='sink')
        c1 = Context('emissions to boot', parent=c)
        d = Context('emulsions', sense='sink')
        self.assertTrue(c1.elementary)
        self.assertFalse(d.elementary)

    def test_parent(self):
        c = Context('emissions', sense='sink')
        d = Context('emissions to air', parent=c)
        e = Context('emissions to urban air', parent=d)
        f = Context('emissions to rural air', 'emissions from high stacks', parent=d)
        self.assertSetEqual(set(i for i in d.subcompartments), {e, f})
        self.assertListEqual([str(k) for k in c.self_and_subcompartments], ['emissions', 'emissions to air',
                                                                            'emissions to rural air',
                                                                            'emissions to urban air'])
        self.assertEqual(e.sense, 'Sink')


class ContextManagerTest(CompartmentContainer.CompartmentManagerTest):
    def _test_class(self, ignore_case=True):
        if ignore_case is False:
            self.skipTest('skipping case sensitive test')
        else:
            return ContextManager()

    def setUp(self):
        self.cm = ContextManager()

    def test_add_from_dict(self):
        self._add_water_dict()
        self.assertEqual(str(self.cm['to water']), 'water emissions')
        self.assertEqual(self.cm['to water'].sense, 'Sink')

    def test_retrieve_by_tuple(self):
        self._add_water_dict()
        w = self.cm['to water']
        self.assertIs(w, self.cm[('emissions', 'water emissions')])

    def test_relative_add(self):
        self._add_water_dict()
        uw = self.cm['to water']
        ud = self.cm.add_compartments(['to water', 'lake water'])
        self.assertIs(uw, ud.parent)
        self.assertListEqual(ud.as_list(), ['Emissions', 'water emissions', 'lake water'])

    def test_merge_inconsistent_sense(self):
        d = [
            {
                "name": "to water",
                "parent": "Emissions"
            },
            {
                "name": "fresh water",
                "synonyms": [
                    "freshwater"
                ],
                "parent": "to water"
            },
            {
                "name": "ground-",
                "parent": "fresh water"
            },
            {
                "name": "from ground",
                "synonyms": [
                    "ground-",
                ],
                "parent": "Resources"
            }
        ]
        for k in d[:3]:
            self.cm._add_from_dict(k)
        with self.assertRaises(InconsistentSense):
            self.cm._add_from_dict(d[3])

    def _add_water_context(self):
        self._add_water_dict()
        c = self.cm.add_compartments(('to water', 'to groundwater'))
        self.cm.add_synonym(c, 'ground-')

    def test_unspecified(self):
        c = self.cm.add_compartments(['emissions', 'water', 'unspecified'])
        self.assertEqual(c.name, 'to water, unspecified')
        self.assertEqual(c.parent.name, 'to water')

    def test_inconsistent_lineage(self):
        self._add_water_context()
        with self.assertRaises(InconsistentLineage):
            self.cm.add_compartments(['resources', 'water', 'ground-'], conflict=None)

    def test_inconsistent_lineage_match(self):
        """
        When an intermediate descendant conflicts, we can either raise the exception (cautious) or do some clever
        regex-based predictive guessing (reckless)
        :return:
        """
        self.skipTest('Too hard to replicate this case in ContextManager')

    def test_inconsistent_lineage_rename(self):
        self._add_water_context()
        c = self.cm.add_compartments(['resources', 'water', 'ground-'], conflict='rename')
        self.assertEqual(c.name, 'from water, ground-')

    def test_inconsistent_lineage_skip(self):
        """
        When an intermediate descendant conflicts, we can either raise the exception (cautious) or do some clever
        regex-based predictive guessing (reckless)
        :return:
        """
        self._add_water_context()
        rw = self.cm.add_compartments(('resources', 'from water'))
        fw = self.cm.add_compartments(['resources', 'water', 'ground-'], conflict='skip')

        self.assertIs(fw, rw)

    def test_elementary(self):
        self.assertEqual(len(list(self.cm.objects)), 2)
        n = self.cm.add_compartments(['elementary flows', 'emissions', 'air', 'urban air'])
        self.assertListEqual(n.as_list(), ['elementary flows', 'Emissions', 'to air', 'urban air'])
        self.assertTrue(n.elementary)
        self.assertIn(self.cm['Emissions'], self.cm['Elementary flows'].subcompartments)
        self.assertEqual(len(list(self.cm.objects)), 5)

    def test_protected(self):
        """
        Protected terms are 'air', 'water', and 'ground'- if these are added as subcompartments to compartments with
        non-None sense, they are modified to e.g. 'to air' or 'from air' appropriate to the sense.
        :return:
        """
        pass

    def test_as_list(self):
        r_l = ['elementary flows', 'resources', 'water', 'subterranean']
        e_l = ['elementary flows', 'emissions', 'water', 'subterranean']
        r = self.cm.add_compartments(r_l)
        e = self.cm.add_compartments(e_l)
        self.assertEqual(r.name, 'subterranean')
        self.assertEqual(e.name, 'to water, subterranean')
        self.assertListEqual(r.as_list(), ['elementary flows', 'Resources', 'from water', 'subterranean'])
        self.assertListEqual(e.as_list(), ['elementary flows', 'Emissions', 'to water', 'subterranean'])


class DefaultContextsTest(unittest.TestCase):
    def setUp(self):
        self.cm = ContextManager(source_file=DEFAULT_CONTEXTS)

    def test_load(self):
        self.assertEqual(len(self.cm), 35)
        self.assertSetEqual({k.name for k in self.cm.top_level_compartments}, {'Emissions', 'Resources'})

    def test_matching_compartment(self):
        foreign_cm = ContextManager()
        fx = foreign_cm.add_compartments(('resources', 'water', 'CA', 'CA-QC'))
        fx.add_origin('dummy.test')
        cx = self.cm.find_matching_context(fx)
        self.assertEqual(cx.sense, 'Source')
        self.assertIs(cx.top(), self.cm['Resources'])
        self.assertListEqual(cx.as_list(), ['Resources', 'from water'])  # superfluous information trimmed
        self.assertIs(self.cm['dummy.test:CA-QC'], cx)

    def test_context_hint(self):
        self.cm.add_context_hint('dummy.test', 'air', 'to air')
        tgt = self.cm['to air']
        self.assertIs(self.cm['dummy.test:air'], tgt)

    def test_matching_sublineage_alt(self):
        """
        This was originally going to test the disregarded functionality- only to realize that does not get triggered
        by find_matching_context.. without that, it is more or less the same as the test below it
        :return:
        """
        self.cm.add_context_hint('dummy.test', 'air', 'to air')
        tgt = self.cm['to air']
        foreign_cm = ContextManager()
        fx = foreign_cm.add_compartments(('Elementary Flows', 'NETL Coal Elementary Flows', 'air', 'unspecified'))
        fx.add_origin('dummy.test')
        self.assertEqual(fx.name, 'air, unspecified')  # just throwing this in there
        self.assertIs(self.cm.find_matching_context(fx), tgt)
        # self.assertIn('Elementary Flows', self.cm.disregarded_terms)

    def test_matching_sublineage(self):
        self.cm.add_context_hint('dummy.test', '[resources]', 'Resources')
        tgt = self.cm['from ground']
        self.assertIs(self.cm['dummy.test:[resources]'], tgt.parent)
        foreign_cm = ContextManager()
        fx = foreign_cm.add_compartments(('Elementary Flows', 'NETL Coal Elementary Flows', 'NETL Elementary Flows',
                                          ' [Resources] ', 'ground'))
        # context gets its origin added when used in a Characterization or an Exchange- for now we do it manually
        fx.add_origin('dummy.test')
        self.assertIs(self.cm.find_matching_context(fx), tgt)
        self.assertIs(self.cm['dummy.test:ground'], tgt)


if __name__ == '__main__':
    unittest.main()


from setuptools import setup, find_packages

requires = [
    "antelope_interface>=0.1.4",
    "xlrd==1.2.0",  ## really need to get out of this requirement and into openpyxl or equivalent
    "python-magic>=0.4.18",
    "requests>=2.25"
]

# optional: pylzma
"""
Version History
0.1.5 - 2021-02-05 - Updates to NullContext handling, flow term matching, fixed faulty requirements (add requests) 
0.1.4 - 2021/01/29 - background passing
0.1.3 - 2021/01/29 - Build passing (without bg tested)
0.1.3rc4           - fix last edit
0.1.3rc3           - xlrd removes support for xlsx in 2.0; _localize_file(None) return None
0.1.3rc2           - include data files
0.1.3rc1           - update requirements

0.1.2b - 2020/12/29 - fix some minor items
0.1.2 - 2020/12/28 - PyPI installation; includes significant performance enhancements for LCIA 

0.1.1 - 2020/11/12 - Bug fixes all over the place.  
                     Catalogs implemented
                     LCIA computation + flat LCIA computation reworked
                     Improvements for OpenLCA LCIA methods

0.1.0 - 2020/07/31 - Initial release - JIE paper
"""

VERSION = '0.1.5'

setup(
    name="antelope_core",
    version=VERSION,
    author="Brandon Kuczenski",
    author_email="bkuczenski@ucsb.edu",
    license="BSD 3-Clause",
    install_requires=requires,
    extras_require={
        'XML': ['lxml>=1.2.0'],
        'write_to_excel': ['xlsxwriter>=1.3.7']
    },
    include_package_data=True,
    url="https://github.com/AntelopeLCA/core",
    summary="A reference implementation of the Antelope interface for accessing a variety of LCA data sources",
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    packages=find_packages()
)
